/*=========================================================================*\
* Hash table implementation
* Diego Nehab
* 21/9/1999
\*=========================================================================*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "hash.h"

/*=========================================================================*\
* Auxiliary functions declaration
\*=========================================================================*/
static char* strdup_(const char *s);
static unsigned long hash_elf(const char *ident);
static long int hash_index(p_hash hash, const char *ident);
static int hash_grow(p_hash hash);

static unsigned long prime_size[] = {
	7, 13, 31, 61, 127, 251, 509, 1021, 2039, 4093, 8191, 16381,
	32749, 65521, 131071, 262139, 524269, 1048573, 2097143
};

#define MAX_STEP ((int) (sizeof(prime_size)/sizeof(unsigned long)))

/*=========================================================================*\
* Exported function definitions
\*=========================================================================*/
/*-------------------------------------------------------------------------*\
* Creates a new hash table
* Returns
*   pointer to newly created hash table in case of success. NULL othrewise
\*-------------------------------------------------------------------------*/
p_hash hash_create(void)
{
	p_hash hash;
	long int c;

	hash = (p_hash) malloc(sizeof(t_hash));
	if (!hash)
		return NULL;

	hash->step = 0;
	hash->size = prime_size[hash->step];

	hash->table = (p_hashnode) malloc(sizeof(t_hashnode) * hash->size);
	hash->index = (long int *) malloc(sizeof(long int) * hash->size);
	if (!hash->table || !hash->index) {
		if (hash->table) free(hash->table);
		if (hash->index) free(hash->index);
		free(hash);
		return NULL;
	}

	/* initialize table entries */
	for (c = 0; c < hash->size; c++) {
		hash->table[c].ident = NULL;
		hash->table[c].num = 0;
		hash->table[c].order = -1;
		hash->index[c] = -1;
	}
	hash->used = 0;

	return hash;
}

/*-------------------------------------------------------------------------*\
* Inserts a new table entry
* Input
*   hash: hash table
*   ident: entry name. a null pointer is ignored
*   data: associated entry data
* Output
*   order: number of elements inserted in the table before the entry. 
*     a null pointer is ignored
* Returns
*   0 if error, 1 if new element, 2 if dupplicate (data is not altered
*   in this case)
* Obs
*   a copy of ident is stored in the table
\*-------------------------------------------------------------------------*/
int hash_insert(p_hash hash, const char *ident, long int *order)
{
	long int index, size, used;
   int err;
	p_hashnode node; 

	/* make sure we don't crash */
	assert(hash && (hash->size > 0) && (hash->used >= 0) && hash->table);
	if (!hash || (hash->size <= 0) || (hash->used < 0) || 
		!hash->table || !ident)
		return 0;

	size = hash->size;
	used = hash->used;
	/* if table is 75% full, try to grow it */
	if (4*used >= 3*size)
		/* if we couldn't grow it and table is full, get out */
		if (!hash_grow(hash) && used >= size)
			return 0;

	/* determine entry position in table */
	index = hash_index(hash, ident);
	node = hash->table + index;

	/* if slot is empty, set it to new node */
	if (!node->ident) {
		node->ident = strdup_(ident);
		node->order = hash->used;
		hash->used++;
		hash->index[node->order] = index;
		err = 1;
	} else
		err = 2;

   node->num++;

	if (order) *order = node->order;

	return err;
}

/*-------------------------------------------------------------------------*\
* Retrieve an entry from the table
* Input
*   hash: hash table
*   ident: entry name
* Output
*   order: number of elements inserted in the table before the entry. 
*     a null pointer is ignored
*   num: number of color occurences
* Returns
*   0 if error or not found, 1 if found
\*-------------------------------------------------------------------------*/
int hash_get(p_hash hash, const char *ident, long int *num, long int *order)
{
	long int index;
	p_hashnode node; 

	/* make sure we don't crash */
	assert(hash && (hash->size > 0) && (hash->used >= 0) && hash->table);
	if (!hash || (hash->size <= 0) || (hash->used < 0) || 
		!hash->table || !ident)
		return 0;

	index = hash_index(hash, ident);
	node = hash->table + index;

	/* slot is empty */
	if (!node->ident) {
		if (order) *order = -1;
		if (num) *num = 0;
	/* found */
	} else {
		if (order) *order = node->order;
		if (num) *num = node->num;
	}

	return 1;
}

/*-------------------------------------------------------------------------*\
* Iterate on table elements in insertion order
* Input
*   hash: hash table
*   callback: function to be called on each element
* Returns
*   1 in case of success, 0 otherwise
\*-------------------------------------------------------------------------*/
int hash_iterate(p_hash hash, p_hashcallback callback)
{
	p_hashnode node;
	long int c;
	int err;

	/* make sure we don't crash */
	assert(hash && (hash->size > 0) && (hash->used >= 0) && 
		hash->table && hash->index && callback);
	if (!hash || (hash->size <= 0) || (hash->used < 0) || 
		!hash->table || !hash->index || !callback)
		return 0;

	for (c = 0; c < hash->used; c++) {
		node = hash->table + hash->index[c];
		err = callback(node->ident, node->order, node->num); 
		if (err != 1)
			return err;
	}

	return 1;
}

/*-------------------------------------------------------------------------*\
* Frees all memory occupied by hash table
* Input
*   hash: hash table
* Returns
*   1 in case of success, 0 otherwise
\*-------------------------------------------------------------------------*/
int hash_kill(p_hash hash)
{
	long int c;

	/* make sure we don't crash */
	assert(hash && (hash->size > 0) && (hash->used >= 0) && 
		hash->table && hash->index);
	if (!hash || (hash->size <= 0) || hash->used < 0 || 
		!hash->table || !hash->index)
		return 0;

	for (c = 0; c < hash->used; c++)
		free(hash->table[hash->index[c]].ident);

	free(hash->index);
	free(hash->table);
	free(hash);

	return 1;
}

/*=========================================================================*\
* Auxiliary functions definition
\*=========================================================================*/
/*-------------------------------------------------------------------------*\
* Returns a hash for a string
* Input
*   ident: string to be hashed
* Returns
*   hash value
\*-------------------------------------------------------------------------*/
static unsigned long int hash_elf(const char *ident)
{
  unsigned long h, g;

  h = 0;
  while (*ident) {
    h = (h << 4) + *ident++;
    if ((g = h & 0xF000000)) {
      h ^= g >> 24;
    }
    h &= ~g;
  }
  return h;
}

/*-------------------------------------------------------------------------*\
* Grow table to next step in geometric progression growth
* Input
*   hash: hash table
* Returns
*   0 if error, 1 if ok
\*-------------------------------------------------------------------------*/
static int hash_grow(p_hash hash)
{
	p_hashnode old_table;
	long int *old_index;
	long int old_used, old_size;
	long int c, oi, ni;

	/* make sure we don't crash */
	assert(hash && (hash->size > 0) && (hash->used >= 0) && 
		hash->table && hash->index);
	if (!hash || (hash->size <= 0) || (hash->used < 0) || 
		!hash->table || !hash->index)
		return 0;

	if (hash->step >= MAX_STEP)
		return 0;

	old_table = hash->table;
	old_index = hash->index;
	old_size = hash->size;
	old_used = hash->used;

    hash->step++;
    hash->size = prime_size[hash->step];
	hash->table = (p_hashnode) malloc(sizeof(t_hashnode) * hash->size);
	hash->index = (long int *) malloc(sizeof(long int) * hash->size);

	if (!hash->table || !hash->index) {
		if (hash->table) free(hash->table);
		if (hash->index) free(hash->index);
		hash->table = old_table;
		hash->index = old_index;
		hash->step--;
		hash->size = old_size;
		hash->used = old_used;
		return 0;
    }

	/* initialize new table */
	for (c = 0; c < hash->size; c++) {
		hash->table[c].ident = NULL;
		hash->table[c].num = 0;
		hash->table[c].order = -1;
		hash->index[c] = -1;
	}
	hash->used = 0;

	/* copy old elements into new table, in the order they were added */
	for (c = 0; c < old_used; c++) {
		oi = old_index[c];
		ni = hash_index(hash, old_table[oi].ident);
		hash->table[ni] =  old_table[oi];
		hash->used++;
		hash->index[c] = ni;
	}

	free(old_table);
	free(old_index);
	return 1;
}

/*-------------------------------------------------------------------------*\
* Find node position in table
* Input
*   ident: node identifier
* Returns
*   index on table
\*-------------------------------------------------------------------------*/
#define rehash(oh, ih, ts) (oh + (ih % (ts-1)) + 1) 
#define different(a, b) (strcmp(a, b) != 0) 
static long int hash_index(p_hash hash, const char *ident)
{
	long int h1, index, size;
	p_hashnode node; 

	size = hash->size;
	h1 = index = hash_elf(ident) % size;
	node = hash->table + index;
	while (node->ident && different(node->ident, ident)) {
		index = rehash(index, h1, size); 
		if (index >= hash->size) index -= hash->size; 
		node = hash->table + index;
	}

	return index;
}

/*-------------------------------------------------------------------------*\
* Allocates memory for and return a copy of a string
* Input
*   s: string to copy
* Returns
*   pointer to copied string
\*-------------------------------------------------------------------------*/
static char* strdup_(const char *s)
{
  char *temp;

  if (s == NULL)
    return NULL;

  temp = malloc(strlen(s) + 1);

  if (temp)
    strcpy(temp, s);

  return temp;
}

/*=========================================================================*\
* Test procedures
\*=========================================================================*/
#ifdef HASH_TEST
int callback(const char *ident, int order, p_combtree data) 
{ printf(ident); return 1; }

int pain(void) 
{
	p_hash hash;
	hash = hash_create();
	hash_insert(hash, "um", NULL, NULL);
	hash_insert(hash, "dois", NULL, NULL);
	hash_insert(hash, "tres", NULL, NULL);
	hash_insert(hash, "quatro", NULL, NULL);
	hash_insert(hash, "cinco", NULL, NULL);
	hash_insert(hash, "seis", NULL, NULL);
	hash_insert(hash, "um", NULL, NULL);
	hash_insert(hash, "um", NULL, NULL);
	hash_insert(hash, "um", NULL, NULL);
	hash_insert(hash, "um", NULL, NULL);
	hash_insert(hash, "um", NULL, NULL);
	hash_insert(hash, "um", NULL, NULL);
	hash_insert(hash, "um", NULL, NULL);
	hash_kill(hash);
}


int main(void)
{
	FILE *fin;
	p_hash hash;
	char word[300];
	int inserted = 0;

	hash = hash_create();
	fin = fopen("/usr/dict/words", "r");
	while ((fgets(word, sizeof(word), fin))) {
		inserted++;
		hash_insert(hash, word, NULL, NULL);
	}
	printf("%d, %d, %d\n", hash->used, inserted, hash->size);
	hash_kill(hash);
	return 1;
}
#endif /* HASH_TEST */

