/** MODIFICADO POR JULIANA **/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "object.h"
#include "algebra.h"


/*****************************************************************************/
/* Definicoes de tipos                                                       */
/*****************************************************************************/
/* constantes de tipos de objetos */
enum ObjectType {
	UNKNOWN,
	SPHERE,
	TRIANGLE,
	BOX
};

/* estrutura de um objeto */
struct Object_impl {
	/* tipo do objeto - um dos valores da enumeracao ObjectType */
	int type;

	/* indice do material do objeto na cena */
	int material;

	/* dado especifico que depende do tipo do objeto */
	void *data;
};

/* definicao dos dados especificos de uma esfera */
typedef struct {
	Vector3D center; 	/* centro da esfera */
	double radius;     	/* raio da esfera */
} Sphere;

/* definicao dos dados especificos de um triangulo */
typedef struct {
	Vector3D v0, v1, v2;	/* vertices do triangulo */
} Triangle;

/* definicao dos dados especificos de uma caixa alinhada com os planos 
   principais */
typedef struct {
	Vector3D ll, ur;	/* ponto inferior esquerdo e superior direito */
} Box;


/*****************************************************************************/
/* Implementacao das funcoes exportadas                                      */
/*****************************************************************************/
Object * objCreateSphere( int mat, Vector3D center, double radius )
{
	Object * obj;
	Sphere * sph;

	/* alocacao de memoria */
	obj = (Object *)malloc(sizeof(Object));
	if (obj == NULL) return NULL;

	sph = (Sphere *)malloc(sizeof(Sphere));
	if (sph == NULL) {
		free(obj);
		return NULL;
	}

	/* inicializacao do objeto */
	sph->center = center;
	sph->radius = radius;
	obj->type = SPHERE;
	obj->material = mat;
	obj->data = (void *) sph;

	return obj;
}

Object * objCreateTriangle( int mat, Vector3D v0, Vector3D v1, Vector3D v2 )
{
	Object * obj;
	Triangle * tri;

	/* alocacao de memoria */
	obj = (Object *)malloc(sizeof(Object));
	if (obj == NULL) return NULL;

	tri = (Triangle *)malloc(sizeof(Triangle));
	if (tri == NULL) {
		free(obj);
		return NULL;
	}

	/* inicializacao do objeto */
	tri->v0 = v0;
	tri->v1 = v1;
	tri->v2 = v2;
	obj->type = TRIANGLE;
	obj->material = mat;
	obj->data = (void *) tri;

	return obj;
}

Object * objCreateBox( int mat, Vector3D ll, Vector3D ur )
{
	Object * obj;
	Box * box;

	/* alocacao de memoria */
	obj = (Object *)malloc(sizeof(Object));
	if (obj == NULL) return NULL;

	box = (Box *)malloc(sizeof(Box));
	if (box == NULL) {
		free(obj);
		return NULL;
	}

	/* inicializacao do objeto */
	box->ll = ll;
	box->ur = ur;
	obj->type = BOX;
	obj->material = mat;
	obj->data = (void *) box;

	return obj;
}

void objDestroy( Object *obj )
{
	/* verifica integridade do objeto */
	assert(obj);
	assert(obj->data);
	assert(obj->type == SPHERE || obj->type == TRIANGLE
	       || obj->type == BOX);

	/* libera memoria alocada */
	free(obj->data);
	free(obj);
}

double objIntercept( Object *obj, Vector3D eye, Vector3D ray )
{
	/* verifica integridade do objeto */
	assert(obj);
	assert(obj->data);
	assert(obj->type == SPHERE || obj->type == TRIANGLE
	       || obj->type == BOX);

	/* executa a intersecao dependendo do tipo do objeto */
	if (obj->type == SPHERE) {
		Sphere *sph = (Sphere *) obj->data;
		
   /* codigo de intersecao de esfera */
      double t      ,  /* distncia do olho */       
             dA     ,  /* coeficientes de dA * t^2 + dB * t + dC = 0 */
             dB     ,
             dC     ,
             dDelta ;
             
      dA = v3dDot( ray, ray ) ; /* dA = ray . ray */

   /* dB = ( 2 * ray ) . ( eye - center ) */
      dB = v3dDot( v3dScale( 2., ray ), v3dSub( eye, sph->center ) ) ; 

   /* dC = ( eye - center ) . ( eye - center ) - radius^2 */
      dC = v3dDot( v3dSub( eye, sph->center ), v3dSub( eye, sph->center ) ) ;
      dC -= sph->radius * sph->radius ; 

      dDelta = dB * dB - 4. * dA * dC ; /* dDelta = dB^2 - 4 * dA * dC */ 
   
      if ( dDelta >= 0 ) {             /* se dDelta < 0, no interceptou */
         
         t = ( - dB - sqrt( dDelta ) ) / ( 2. * dA ) ;
         
      /* se t >= 1, j achou ponto mais prximo da interseo com a esfera */
         if ( t < 1. ) {
            t = ( - dB + sqrt( dDelta ) ) / ( 2. * dA ) ; 
		   } /* if */

      } else {
         return -1. ;   /* dDelta < 0, no interceptou */
      } /* if */

   		return t ;
		
	} else if (obj->type == TRIANGLE) {
		Triangle *tri = (Triangle *) obj->data;
		
   /* codigo de intersecao do triangulo */
      double   t            , /* distncia do olho */
               dt1pontot2   , /* t1 . t2 */  
               dt1pontot3   ; /* t1 . t3 */
      Vector3D NormPlanoTri , /* normal do plano do tringulo */
               CompPlano1   , /* um vetor que define o plano */
               CompPlano2   , /* outro vetor que define o plano */
               CompPlano3   , /* um vetor no plano do tringulo */
               PtPlano      , /* ponto do plano do tringulo */
               t1           , /* vetores auxiliares */
               t2           ,
               t3           ;

   /* acha 2 vetores que definem o plano do tringulo */
      CompPlano1 = v3dSub( tri->v1, tri->v0 ) ;
      CompPlano2 = v3dSub( tri->v2, tri->v0 ) ;

   /* acha normal do plano do tringulo */
      NormPlanoTri = v3dCross( CompPlano1, CompPlano2 ) ;   
      
   /* calcula t = ( ( tri->v0 - eye ) . NormPlanoTri ) 
                / ( ray . NormPlanoTri ) */
      t = v3dDot( v3dSub( tri->v0, eye ), NormPlanoTri ) ;
      t /= v3dDot( ray, NormPlanoTri ) ;

   /* calcula ptPlano = eye + t * ray */
      PtPlano = v3dAdd( eye, v3dScale( t, ray ) ) ;
   
   /* acha vetores do plano do tringulo */   
      CompPlano3 = v3dMinus( CompPlano2 ) ;
      CompPlano2 = v3dSub( tri->v2, tri->v1 ) ;

   /* acha vetores para saber se P pertence ao tringulo */
      t1 = v3dCross( CompPlano1, v3dSub( PtPlano, tri->v0 ) ) ;
      t2 = v3dCross( CompPlano2, v3dSub( PtPlano, tri->v1 ) ) ;
      t3 = v3dCross( CompPlano3, v3dSub( PtPlano, tri->v2 ) ) ; 

	   dt1pontot2 = v3dDot( t1, t2 ) ; /* calcula t1 . t2 */
      dt1pontot3 = v3dDot( t1, t3 ) ; /* calcula t1 . t3 */
      
   /* se estes produtos > 0, t1, t2 e t3 tm o mesmo sentido e P
      est dentro do tringulo */
      if ( dt1pontot2 > 0
        && dt1pontot3 > 0 ) {
         return t ;
      } else {
         return -1. ;
      } /* if */

	} else if (obj->type == BOX) {
		Box *box = (Box *) obj->data;
		
   /* codigo de intersecao de box */
      double t  ,
             dX ,
             dY ,
             dZ ;

   /* v se componente x da ray  != 0 */
      if ( ray.x != 0. ) {

      /* se ray.x > 0, s se v Xmin da caixa, seno, s Xmax */
         if ( ray.x > 0. ) {
            t = ( box->ll.x - eye.x ) / ray.x ;
         } else {
            t = ( box->ur.x - eye.x ) / ray.x ;
         } /* if */

      /* tendo t, calcula dY e dX */
         dY = eye.y + t * ray.y ;
         dZ = eye.z + t * ray.z ;

      /* testa se dY e dZ esto na caixa */
         if ( dY >= box->ll.y
           && dY <= box->ur.y
           && dZ >= box->ll.z
           && dZ <= box->ur.z ) {
            return t ;
         } /* if */

      } /* if */

   /* v se componente y da ray  != 0 */
      if ( ray.y != 0. ) {

      /* se ray.y > 0, s se v Ymin da caixa, seno, s Ymax */
         if ( ray.y > 0. ) {
            t = ( box->ll.y - eye.y ) / ray.y ;
         } else {
            t = ( box->ur.y - eye.y ) / ray.y ;
         } /* if */

      /* tendo t, calcula dX e dZ */
         dX = eye.x + t * ray.x ;
         dZ = eye.z + t * ray.z ;

      /* testa se dZ e dX esto na caixa */
         if ( dX >= box->ll.x
           && dX <= box->ur.x
           && dZ >= box->ll.z
           && dZ <= box->ur.z ) {
            return t ;
         } /* if */

      } /* if */

   /* v se componente z da ray  != 0 */
      if ( ray.z != 0. ) {

      /* se ray.z > 0, s se v Zmin da caixa, seno, s Zmax */
         if ( ray.z > 0. ) {
            t = ( box->ll.z - eye.z ) / ray.z ;
         } else {
            t = ( box->ur.z - eye.z ) / ray.z ;
         } /* if */

      /* tendo t, calcula dY e dZ */
         dY = eye.y + t * ray.y ;
         dX = eye.x + t * ray.x ;

      /* testa se dY e dX esto na caixa */
         if ( dY >= box->ll.y
           && dY <= box->ur.y
           && dX >= box->ll.x
           && dX <= box->ur.x ) {
            return t ;
         } /* if */

      } /* if */

		return -1 ;  /* no houve interseo com a caixa */

	} else {
		/* nunca deve acontecer - tipo invalido de objeto */
		return (-1.0);
	}
}

Vector3D objGetNormal( Object *obj, Vector3D pos )
{
	/* verifica integridade do objeto */
	assert(obj);
	assert(obj->data);
	assert(obj->type == SPHERE || obj->type == TRIANGLE
	       || obj->type == BOX);

	/* calcula a normal dependendo do tipo do objeto */
	if (obj->type == SPHERE) {
		Sphere *sph = (Sphere *) obj->data;

		/* codigo de calculo da normal da esfera no ponto dado */
		return v3dSub( pos, sph->center );
	} else if (obj->type == TRIANGLE) {
		Triangle *tri = (Triangle *) obj->data;

		/* codigo de calculo da normal do triangulo */
		return v3dCross(
			v3dSub(tri->v1, tri->v0),
			v3dSub(tri->v2, tri->v0)
		);
	} else if (obj->type == BOX) {
		Box *box = (Box *) obj->data;

      /* codigo a normal da Box */
   
   /* testa se P pertence face X = Xmin */
      if ( fabs( pos.x - box->ll.x ) < D_ZERO ) {
         return v3dCreate( -1, 0, 0 ) ;
      } /* if */

   /* testa se P pertence face X = Xmax */
      if ( fabs( pos.x - box->ur.x ) < D_ZERO ) {
         return v3dCreate( 1, 0, 0 ) ;
      } /* if */

   /* testa se P pertence face Y = Ymin */
      if ( fabs( pos.y - box->ll.y ) < D_ZERO ) {
         return v3dCreate( 0, -1, 0 ) ;
      } /* if */

   /* testa se P pertence face Y = Ymax */
      if ( fabs( pos.y - box->ur.y ) < D_ZERO ) {
         return v3dCreate( 0, 1, 0 ) ;
      } /* if */

   /* testa se P pertence face Z = Zmin */
      if ( fabs( pos.z - box->ll.z ) < D_ZERO )  {
         return v3dCreate( 0, 0, -1 ) ;
      } /* if */

   /* testa se Z pertence face Z = Zmax */
      if ( fabs( pos.z - box->ur.z ) < D_ZERO ) {
         return v3dCreate( 0, 0, 1 ) ;
      } /* if */

	} /* if */

   /* nunca deve acontecer - tipo invalido de objeto */
	return v3dCreate(0,0,0);
}

Object * objReadSphere( FILE *file )
{
	Vector3D center;
	double radius;
	int matno;

	fscanf(file, "%i", &matno);
	fscanf(file, "%lf", &radius);
	center = v3dRead(file);

#ifndef NDEBUG
	printf("\n[%s:%d] read sphere:\n", __FILE__, __LINE__);
	printf("\tmaterial = %i,\n", matno);
	printf("\tradius = %g,\n", radius);
	printf("\tcenter = (%g, %g, %g)\n", center.x, center.y, center.z);
#endif

	return objCreateSphere( matno, center, radius );
}

Object * objReadTriangle( FILE *file )
{
	Vector3D v0, v1, v2;
	int matno;

	fscanf(file, "%i", &matno);
	v0 = v3dRead(file);
	v1 = v3dRead(file);
	v2 = v3dRead(file);

#ifndef NDEBUG
	printf("\n[%s:%d] read triangle:\n", __FILE__, __LINE__);
	printf("\tmaterial = %i,\n", matno);
	printf("\tv0 = (%g, %g, %g),\n", v0.x, v0.y, v0.z);
	printf("\tv1 = (%g, %g, %g),\n", v1.x, v1.y, v1.z);
	printf("\tv2 = (%g, %g, %g)\n", v2.x, v2.y, v2.z);
#endif

	return objCreateTriangle( matno, v0, v1, v2 );
}

Object * objReadBox( FILE *file )
{
	Vector3D ll, ur;
	int matno;
   double dXmin , /* coordenadas mnimas e mximas da caixa */
          dXmax ,
          dYmin ,
          dYmax ,
          dZmin ,
          dZmax ;

	fscanf(file, "%i", &matno);
	ll = v3dRead(file);
	ur = v3dRead(file);

/* acha coordenadas mximas e mnimas da caixa */
   dXmin = ll.x < ur.x ? ll.x : ur.x ;
   dXmax = ll.x > ur.x ? ll.x : ur.x ;
   dYmin = ll.y < ur.y ? ll.y : ur.y ;
   dYmax = ll.y > ur.y ? ll.y : ur.y ;
   dZmin = ll.z < ur.z ? ll.z : ur.z ;
   dZmax = ll.z > ur.z ? ll.z : ur.z ;

   ll = v3dCreate( dXmin, dYmin, dZmin ) ;  /* acha canto inferior esquerdo */
   ur = v3dCreate( dXmax, dYmax, dZmax ) ;  /* acha canto superior direito */

#ifndef NDEBUG
	printf("\n[%s:%d] read box:\n", __FILE__, __LINE__);
	printf("\tmaterial = %i,\n", matno);
	printf("\tll = (%g, %g, %g),\n", ll.x, ll.y, ll.z);
	printf("\tur = (%g, %g, %g)\n", ur.x, ur.y, ur.z);
#endif

	return objCreateBox( matno, ll, ur );
}


/*==============================================================================
*Nome: iPegarIdxMat
*=============================================================================*/

int iPegarIdxMat( Object* pObjeto ) {
   return pObjeto->material ;
} /* iPegarIdxMat */