/** MODIFICADO POR JULIANA **/

#include "scene.h"

#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

/* numero maximo de materiais por cena */
#define MAX_MATERIALS	64

/* numero maximo de objetos por cena */
#define MAX_OBJECTS	64

/* numero maximo de luzes por cena */
#define MAX_LIGHTS	8

struct Scene_impl {
	/* camera da cena */
	Camera *camera;
	
	/* cor do fundo da cena - soh usado se 'bgimg' for NULL */
	Color bgcolor;

	/* imagem usada como fundo da cena ou NULL */
	Image *bgimg;

	/* tabela de materiais usados na cena */
	Material mat_tbl[MAX_MATERIALS];

	/* numero de materiais usados na cena (entre 0 e MAX_MATERIALS) */
	int mat_tbl_size;

	/* tabela de objetos da cena */
	Object *obj_tbl[MAX_OBJECTS];

	/* numero de objetos da cena (entre 0 e MAX_OBJECTS) */
	int obj_tbl_size;

	/* tabela de luzes da cena */
	Light light_tbl[MAX_LIGHTS];

	/* numero de luzes na cena (entre 0 e MAX_LIGHTS) */
	int light_tbl_size;
};

int sceGetObjectCount( Scene *scene )
{
	return scene->obj_tbl_size;
}

Object * sceGetObject( Scene *scene, int idx )
{
	assert( idx >= 0 );
	assert( idx < scene->obj_tbl_size );

	return scene->obj_tbl[idx];
}

int sceGetLightCount( Scene *scene )
{
	return scene->light_tbl_size;
}

Light sceGetLight( Scene *scene, int idx )
{
	assert( idx >= 0 );
	assert( idx < scene->light_tbl_size );

	return scene->light_tbl[idx];
}

int sceGetMaterialCount( Scene *scene )
{
	return scene->mat_tbl_size;
}

Material sceGetMaterial( Scene *scene, int idx )
{
	assert( idx >= 0 );
	assert( idx < scene->mat_tbl_size );

	return scene->mat_tbl[idx];
}

/* findCategory - Procura a proxima 'categoria' de dados no arquivo .rt2 e
   coloca seu nome em 'buf'. Retorna 0 se encontrar um erro ou 1 se tiver
   sucesso. */
static int findCategory( FILE *fp, char *buf, int bufsize )
{
	int c, i, len;
	
	do {
		c = fgetc(fp);
		if (c == EOF) return 0;
	} while (isspace(c));

	/* nao eh inicio de categoria, sao mais dados */
	if (c != '#') {
		ungetc(c, fp);
		return 0;
	}

	/* le nome da categoria */
	if (fgets( buf, bufsize, fp ) == NULL) return 0;

	/* converte para caixa alta e tira '\n' do final */
	len = strlen( buf );
	
	for (i=0; i<len; i++) {
		if (buf[i] == '\n' || buf[i] == '\r') {
			/* termina a string no primeiro '\n' */
			buf[i] = 0;
			return 1;
		} else if (isalpha(buf[i])) {
			buf[i] = toupper(buf[i]);
		} else {
			/* caractere invalido encontrado na categoria */
			/* retorna com erro */
			return 0;
		}
	}

	return 1;
}

/* skipCategory - Pula todos os dados ateh encontrar a proxima categoria.
   Os parametros sao tratados como em 'findCategory'. */
static int skipCategory( FILE *fp, char *buf, int bufsize )
{
	int c;
	
	do {
		c = fgetc(fp);
		if (c == EOF) return 0;
	} while (c != '#');

	ungetc( c, fp );
	return findCategory( fp, buf, bufsize );
}

/* preprocess - Pre-processa o arquivo de dados 'fp' gerando um novo arquivo
   temporario que nao contem os comentarios. Retorna NULL em caso de erro.
   O arquivo original serah fechado por esta operacao. */
static FILE * preprocess( FILE *fp )
{
	char buffer[1024];
	FILE *tmp;

	fscanf( fp, "%s", buffer );
	if (strcmp(buffer, "RT") != 0) return NULL;

	fscanf( fp, "%s", buffer );
	if (strcmp(buffer, "2.0") != 0) return NULL;

	tmp = tmpfile();
	if (tmp == NULL) return NULL;


	while (!feof(fp)) {
		char *p;

		/* le uma linha do arquivo original */
		if (fgets( buffer, 1024, fp ) == NULL) break;

		/* termina string no inicio do comentario */
		/* tira espacos em branco antes do inicio do comentario */
		p = strchr( buffer, '!' );
		if (p) {
			while ( p > buffer && isspace(*(p-1)) ) p--;
			*p = 0;
		}

		/* escreve a nova linha no arquivo temporario */
		fputs( buffer, tmp );
		if( buffer[ strlen(buffer)-1 ] != '\n' ) fputs("\n", tmp);
	}

	fclose(fp);
	rewind(tmp);
	return tmp;
}

Scene * sceRead( const char *filename )
{
	char cat[512];
	FILE *fp;
	Scene *sce;
	
	fp = fopen(filename, "rt");
	if (fp == NULL) return NULL;

	fp = preprocess(fp);
	if (fp == NULL) return NULL;

	sce = (Scene *) malloc( sizeof(Scene) );
	if (sce == NULL) return NULL;

	/* inicializacao do objeto cena vazio */
	sce->camera = NULL;
	sce->bgimg = NULL;
	sce->obj_tbl_size = 0;
	sce->light_tbl_size = 0;
	sce->mat_tbl_size = 0;

	while (!feof(fp)) {
		if (!findCategory(fp, cat, 512) && feof(fp)) break;

		if (strcmp(cat, "CAMERA") == 0) {
			assert( sce->camera == NULL );

			/* le camera da cena */
			sce->camera = camRead( fp );
		} else if (strcmp(cat, "SCENE") == 0) {
			char fname[1024];
			char *p;

			assert( sce->bgimg == NULL );

			/* le cor de fundo da cena */
			sce->bgcolor = corRead(fp);

			/* le imagem de fundo da cena */
			fscanf( fp, "%s", fname );

			/* retira quebras de linha */
			do {
				p = strchr( fname, '\r' );
				if (p) *p = 0;
			} while (p);

			do {
				p = strchr( fname, '\n' );
				if (p) *p = 0;
			} while (p);

			if (strcmp(fname, "null") == 0) {
				sce->bgimg = NULL;
			} else {
				sce->bgimg = imgReadPPM( fname );
			}
		} else if (strcmp(cat, "MATERIAL") == 0) {
			/* adiciona novo material `a lista da cena */
			assert( sce->mat_tbl_size < MAX_MATERIALS );

			sce->mat_tbl[ sce->mat_tbl_size ] = matRead( fp );
			sce->mat_tbl_size++;
		} else if (strcmp(cat, "LIGHT") == 0) {
			/* adiciona nova luz `a lista da cena */
			assert( sce->light_tbl_size < MAX_LIGHTS );

			sce->light_tbl[ sce->light_tbl_size ] = lightRead( fp );
			sce->light_tbl_size++;
		} else if (strcmp(cat, "SPHERE") == 0) {
			/* adiciona novo objeto `a lista da cena */
			assert( sce->obj_tbl_size < MAX_OBJECTS );

			sce->obj_tbl[ sce->obj_tbl_size ] = objReadSphere(fp);
			sce->obj_tbl_size++;
		} else if (strcmp(cat, "TRIANGLE") == 0) {
			/* adiciona novo objeto `a lista da cena */
			assert( sce->obj_tbl_size < MAX_OBJECTS );

			sce->obj_tbl[ sce->obj_tbl_size ] = objReadTriangle(fp);
			sce->obj_tbl_size++;
		} else if (strcmp(cat, "BOX") == 0) {
			/* adiciona novo objeto `a lista da cena */
			assert( sce->obj_tbl_size < MAX_OBJECTS );

			sce->obj_tbl[ sce->obj_tbl_size ] = objReadBox(fp);
			sce->obj_tbl_size++;
		} else {
			/* categoria desconhecida, ignora */
			printf("\n[%s:%i] skipping unknown category '#%s'\n",
			       __FILE__, __LINE__, cat);

			if (!skipCategory( fp, cat, 512 )) break;
		}
	}

	fclose(fp);
	return sce;
}

void sceDestroy( Scene *sce )
{
	int i;

	assert( sce != NULL );

	if (sce->camera != NULL) camDestroy( sce->camera );
	if (sce->bgimg != NULL) imgDestroy( sce->bgimg );

	for(i=0; i < sce->obj_tbl_size; i++) {
		objDestroy( sce->obj_tbl[i] );
	}

	for(i=0; i< sce->mat_tbl_size; i++) {
		if (sce->mat_tbl[i].img != NULL) {
			imgDestroy( sce->mat_tbl[i].img );
		}
	}

	free( sce );
}


/*==============================================================================
*Nome: PegarCamera
*=============================================================================*/

Camera* PegarCamera( Scene* pCenario ) {
   return pCenario->camera ;
} /* PegarCamera */


/*==============================================================================
*Nome: PegarCorFundo
*=============================================================================*/

Color PegarCorFundo( Scene* pCenario ) {
   return pCenario->bgcolor ;
} /* PegarCorFundo */
